use actix_web::{http::header::LOCATION, web, HttpResponse, Responder};
use sqlx::PgPool;

use crate::{
    endpoints::{events::NewOrEditEventForm, IdPath},
    utils::{self, ApplicationError},
};
use brass_db::{
    models::{Assignment, AssignmentChangeset, Event, EventChangeset, EventContext, Role, User},
    validation::AsyncValidate,
};

#[actix_web::post("/events/{id}/edit")]
pub async fn post(
    user: web::ReqData<User>,
    pool: web::Data<PgPool>,
    form: web::Form<NewOrEditEventForm>,
    path: web::Path<IdPath>,
) -> Result<impl Responder, ApplicationError> {
    if user.role != Role::Admin && user.role != Role::AreaManager {
        return Err(ApplicationError::Unauthorized);
    }

    let Some(event) = Event::read_by_id_including_location(pool.get_ref(), path.id).await? else {
        return Ok(HttpResponse::NotFound().finish());
    };

    let changeset = EventChangeset {
        amount_of_posten: form.amount,
        clothing: form.clothing,
        location_id: form.location,
        time: (form.date.and_time(form.start), form.end),
        name: form.name.clone(),
        note: form
            .note
            .clone()
            .and_then(|n| if !n.is_empty() { Some(n) } else { None }),
        voluntary_wachhabender: form.voluntarywachhabender,
        voluntary_fuehrungsassistent: form.voluntaryfuehrungsassistent,
    };

    let context = EventContext {
        event: Some(event.id),
        pool: pool.get_ref(),
        user: &user.into_inner(),
    };

    if let Err(e) = changeset.validate_with_context(&context).await {
        return Ok(HttpResponse::UnprocessableEntity().body(e.to_string()));
    };

    if event.start != changeset.time.0 || event.end != changeset.time.1 {
        let assignments_for_event = Assignment::read_all_by_event(pool.get_ref(), event.id).await?;
        for a in assignments_for_event {
            let c = AssignmentChangeset {
                function: a.function,
                time: changeset.time,
            };
            Assignment::update(pool.get_ref(), a.event_id, a.availability_id, c).await?;
        }
    }

    Event::update(pool.get_ref(), event.id, changeset).await?;

    let url = utils::get_return_url_for_date(&form.date);

    Ok(HttpResponse::Found()
        .insert_header((LOCATION, url.clone()))
        .insert_header(("HX-LOCATION", url))
        .finish())
}
