use std::{
    future::{ready, Ready},
    str::from_utf8,
};

use actix_web::{
    body::BoxBody,
    dev::{forward_ready, Service, ServiceRequest, ServiceResponse, Transform},
    Error, HttpMessage,
};
use futures_util::future::LocalBoxFuture;
use tracing_actix_web::RequestId;

pub struct ErrorAppender;

impl<S> Transform<S, ServiceRequest> for ErrorAppender
where
    S: Service<ServiceRequest, Response = ServiceResponse, Error = Error>,
    S::Future: 'static,
{
    type Response = ServiceResponse;
    type Error = Error;
    type InitError = ();
    type Transform = ErrorAppenderMiddleware<S>;
    type Future = Ready<Result<Self::Transform, Self::InitError>>;

    fn new_transform(&self, service: S) -> Self::Future {
        ready(Ok(ErrorAppenderMiddleware { service }))
    }
}

pub struct ErrorAppenderMiddleware<S> {
    service: S,
}

impl<S> Service<ServiceRequest> for ErrorAppenderMiddleware<S>
where
    S: Service<ServiceRequest, Response = ServiceResponse, Error = Error>,
    S::Future: 'static,
{
    type Response = ServiceResponse;
    type Error = Error;
    type Future = LocalBoxFuture<'static, Result<Self::Response, Self::Error>>;

    forward_ready!(service);

    fn call(&self, req: ServiceRequest) -> Self::Future {
        let id = req.extensions().get::<RequestId>().copied();

        let fut = self.service.call(req);

        Box::pin(async move {
            let res = fut.await?;
            if res.status().is_server_error() {
                let (req, res) = res.into_parts();
                let (res, body) = res.into_parts();

                let body_bytes = actix_web::body::to_bytes(body).await.ok().unwrap();
                let mut s = from_utf8(&body_bytes).unwrap().to_string();

                let id_str = if let Some(id) = id {
                    id.to_string()
                } else {
                    String::new()
                };

                s.push_str(" // request-id: ");
                s.push_str(&id_str);

                let res = res.set_body(BoxBody::new(s));
                return Ok(ServiceResponse::new(req, res));
            }

            Ok(res)
        })
    }
}
